from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import dataclass_json

from .badge import Badge
from .fields import default
from .score_stats import ScoreStats


@dataclass_json
@dataclass
class Player:
    """Player data from ScoreSaber"""

    id: str
    name: str = default()
    profile_picture: str = default("profilePicture")
    country: str = default()
    pp: float = default()
    rank: int = default()
    country_rank: Optional[int] = default("countryRank")
    role: Optional[str] = default()
    badges: Optional[List[Badge]] = default()
    histories: str = default()
    score_stats: Optional[ScoreStats] = default("scoreStats")
    permissions: Optional[int] = default()
    banned: Optional[int] = default()
    inactive: Optional[int] = default()



from dataclasses import dataclass

from dataclasses_json import dataclass_json

from .fields import default
from .leaderboard_info import LeaderboardInfo
from .score import Score


@dataclass_json
@dataclass
class PlayerScore:
    """Player score data from ScoreSaber"""

    score: Score = default()
    leaderboard: LeaderboardInfo = default()



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from dataclasses_json import dataclass_json, LetterCase

from .difficulty import Difficulty
from .fields import default, datetime_field
from .score import Score


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class LeaderboardInfo:
    """Leaderboard info data from ScoreSaber"""

    id: int
    song_hash: str
    song_name: str
    song_sub_name: str
    song_author_name: str
    level_author_name: str
    difficulty: Difficulty
    max_score: int
    ranked: bool
    qualified: bool
    loved: bool
    stars: float
    positive_modifiers: bool
    plays: int
    daily_plays: int
    cover_image: str
    player_score: Optional[Score]
    difficulties: Optional[List[Difficulty]]
    max_pp: int = default("maxPP")
    created_date: datetime = datetime_field("createdDate")
    ranked_date: Optional[datetime] = datetime_field("rankedDate")
    qualified_date: Optional[datetime] = datetime_field("qualifiedDate")
    loved_date: Optional[datetime] = datetime_field("lovedDate")



from typing import *

from faker.providers import BaseProvider

from .models import Player, ScoreStats, Score, GameMode, BeatmapDifficulty, Badge


class ScoreSaberProvider(BaseProvider):
    def role(self) -> Optional[str]:
        roles = [
            "Owner", "Nomination Assessment Team", "Ranking Team", "Quality Assurance Team",
            "Ranking Team Recruit", "Criteria Assurance Team", "Supporter", None
        ]

        return self.random_choices(roles, 1)

    def rank(self) -> int:
        return self.random_int(1, 1000000000)

    def history(self) -> str:
        ranks = []

        for index in range(48):
            ranks.append(str(self.rank()))

        return ",".join(ranks)

    def badge(self) -> Badge:
        badge = Badge()

        badge.image = self.generator.image_url(80, 30)
        badge.description = self.generator.sentence()

        return badge

    def badges(self) -> List[Badge]:
        badges = []

        rnd = self.random_digit()

        if rnd == 0:
            return badges

        for index in range(rnd):
            badges.append(self.badge())

        return badges

    def player_basic(self, player_id: int) -> Player:
        player = Player()

        player.player_id = player_id
        player.player_name = self.generator.user_name()
        player.avatar = self.generator.image_url(184, 184)
        player.rank = self.rank()
        player.country_rank = self.rank()
        player.pp = float(self.numerify("%####.##"))
        player.country = self.generator.country_code()
        player.role = self.role()
        player.badges = self.badges()
        player.history = self.history()
        player.permissions = self.random_int(0, 100)
        player.inactive = self.random_int(0, 1)
        player.banned = self.random_int(0, 1)

        return player

    def score_stats(self) -> ScoreStats:
        score_stats = ScoreStats()

        score_stats.total_score = self.random_int(0, 999999999999)
        score_stats.total_ranked_score = self.random_int(0, 999999999999)
        score_stats.average_ranked_accuracy = float(self.numerify("%#.##############"))
        score_stats.total_play_count = self.random_int(0, 99999)
        score_stats.ranked_play_count = self.random_int(0, 99999)

        return score_stats

    def player_full(self, player_id: int) -> Player:
        player = self.player_basic(player_id)

        player.score_stats = self.score_stats()

        return player

    def score(self) -> Score:
        score = Score()

        score.rank = self.rank()
        score.score_id = self.random_int(0, 999999999999)
        score.score = self.random_int(0, 999999999999)
        score.unmodified_score = score.score
        score.mods = self.random_elements(["NF", "FS", "SS", "GN"], unique=True)
        score.pp = float(self.numerify("%##.###"))
        score.weight = float(self.numerify("%#.##############"))
        score.time_set = self.generator.past_datetime()
        score.leaderboard_id = self.random_int(0, 999999999999)
        score.song_hash = self.generator.sha1(raw_output=False)
        score.song_name = self.generator.text(35)
        score.song_sub_name = self.generator.text(35)
        score.song_author_name = self.generator.user_name()
        score.level_author_name = self.generator.user_name()
        score.characteristic = self.random_choices(list(gamemode), 1)
        score.difficulty = self.random_choices(list(BeatmapDifficulty), 1)
        score.max_score = score.score * float(self.numerify("%.###"))

        return score

    def scores(self, count: int = 1) -> List[Score]:
        scores = []

        for _ in range(count):
            scores.append(self.score())

        return scores



from dataclasses import dataclass

from dataclasses_json import dataclass_json

from .fields import default


@dataclass_json
@dataclass
class ScoreStats:
    """Player score stats data from Score Saber"""

    total_score: int = default("totalScore")
    total_ranked_score: int = default("totalRankedScore")
    average_ranked_accuracy: float = default("averageRankedAccuracy")
    total_play_count: int = default("totalPlayCount")
    ranked_play_count: int = default("rankedPlayCount")
    replays_watched: int = default("replaysWatched")



from dataclasses import dataclass

from dataclasses_json import dataclass_json, LetterCase


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Metadata:
    total: int
    page: int
    items_per_page: int



import asyncio
import logging
from asyncio import AbstractEventLoop
from typing import *

from faker import Faker
from outcache import CacheAsync

from .http_client import HttpClient
from .models import *
from .scoresaber_provider import ScoreSaberProvider

_logger = logging.getLogger(__name__)


class ScoreSaber:
    _url = "https://scoresaber.com/api"

    def __init__(self, loop: Optional[AbstractEventLoop] = None, test_mode: bool = False):
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        self.test_mode = test_mode

        self._http_client = HttpClient(self.loop)

        if test_mode:
            _logger.info("Running in test mode")
            self.faker = Faker()
            Faker.seed(76561198283584459)
            self.faker.add_provider(ScoreSaberProvider)

    async def start(self):
        await self._http_client.start()

    async def close(self):
        await self._http_client.close()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def ws_start(self):
        await self._http_client.ws_connect()

    async def ws_close(self):
        await self._http_client.ws_close()

    # /ws

    async def websocket(self) -> AsyncIterable[Union[PlayerScore]]:
        async for message in self._http_client.ws_listen():
            if message["commandName"] == "score":
                yield PlayerScore.from_dict(message["commandData"])

    # /leaderboard

    @CacheAsync(hours=1)
    async def leaderboards(self,
        search: Optional[str] = None,
        ranked: Optional[bool] = None,
        qualified: Optional[bool] = None,
        loved: Optional[bool] = None,
        min_star: Optional[int] = None,
        max_star: Optional[int] = None,
        category: Optional[Category] = None,
        sort: Optional[Sort] = None,
        unique: Optional[bool] = None,
        page: Optional[int] = None
    ) -> LeaderboardInfoCollection:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        url = f"{self._url}/leaderboards"
        params = {
            "search": search, "ranked": ranked, "qualified": qualified,
            "loved": loved, "minStar": min_star, "maxStar": max_star,
            "category": category, "sort": sort, "unique": unique, "page": page
        }

        return await self._http_client.get(LeaderboardInfoCollection, url, params)

    @CacheAsync(hours=1)
    async def leaderboard_info_by_id(self,
        leaderboard_id: int
    ) -> LeaderboardInfo:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        return await self._http_client.get(LeaderboardInfo, f"{self._url}/leaderboard/by-id/{leaderboard_id}/info")

    @CacheAsync(hours=1)
    async def leaderboard_info_by_hash(self,
        beatmap_hash: str,
        difficulty: BeatmapDifficulty,
        game_mode: Optional[GameMode] = None
    ) -> LeaderboardInfo:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        url = f"{self._url}/leaderboard/by-hash/{beatmap_hash}/info"
        params = {
            "difficulty": difficulty, "gameMode": game_mode
        }

        return await self._http_client.get(LeaderboardInfo, url, params)

    @CacheAsync(hours=1)
    async def leaderboard_scores_by_id(self,
        leaderboard_id: int,
        countries: Optional[str] = None,
        search: Optional[str] = None,
        page: Optional[int] = None
    ) -> ScoreCollection:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        url = f"{self._url}/leaderboard/by-id/{leaderboard_id}/scores"
        params = {
            "countries": countries, "search": search, "page": page
        }

        return await self._http_client.get(ScoreCollection, url, params)

    @CacheAsync(hours=1)
    async def leaderboard_scores_by_hash(self,
        beatmap_hash: str,
        difficulty: BeatmapDifficulty,
        countries: Optional[str] = None,
        search: Optional[str] = None,
        page: Optional[int] = None,
        game_mode: Optional[GameMode] = None
    ) -> ScoreCollection:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        url = f"{self._url}/leaderboard/by-hash/{beatmap_hash}/scores"
        params = {
            "difficulty": difficulty, "countries": countries, "search": search,
            "page": page, "gameMode": game_mode
        }

        return await self._http_client.get(ScoreCollection, url, params)

    @CacheAsync(hours=1)
    async def leaderboard_difficulties(self,
        beatmap_hash: str
    ) -> List[Difficulty]:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        return await self._http_client.get(List[Difficulty], f"{self._url}/leaderboard/get-difficulties/{beatmap_hash}")

    # /player

    @CacheAsync(minutes=10)
    async def players(self,
        search: Optional[str] = None,
        page: Optional[int] = None,
        countries: Optional[str] = None
    ) -> PlayerCollection:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        url = f"{self._url}/players"
        params = {
            "search": search, "page": page, "countries": countries
        }

        return await self._http_client.get(PlayerCollection, url, params)

    @CacheAsync(hours=1)
    async def players_count(self,
        search: Optional[str] = None,
        countries: Optional[str] = None
    ) -> int:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        url = f"{self._url}/players/count"
        params = {
            "search": search, "countries": countries
        }

        return await self._http_client.get(int, url, params)

    @CacheAsync(minutes=1)
    async def player_basic(self,
        player_id: int
    ) -> Player:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        return await self._http_client.get(Player, f"{self._url}/player/{player_id}/basic")

    @CacheAsync(minutes=1)
    async def player_full(self,
        player_id: int
    ) -> Player:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        return await self._http_client.get(Player, f"{self._url}/player/{player_id}/full")

    @CacheAsync(minutes=1)
    async def player_scores(self,
        player_id: int,
        limit: Optional[int] = None,
        sort: Optional[ScoreSort] = None,
        page: Optional[int] = None
    ) -> PlayerScoresCollection:
        if self.test_mode:
            raise RuntimeError("Not implemented!")

        url = f"{self._url}/player/{player_id}/scores"
        params = {
            "limit": limit, "sort": sort, "page": page
        }

        return await self._http_client.get(PlayerScoresCollection, url, params)

    # /user

    # TODO: Implement GET /user/{player_id}/refresh
    # TODO: Implement GET /user/@me
    # TODO: Implement GET /user/quest-key

    # /ranking

    # TODO: Implement GET /ranking/requests/top
    # TODO: Implement GET /ranking/requests/belowTop
    # TODO: Implement GET /ranking/request/{request_id}
    # TODO: Implement GET /ranking/request/by-id/{leaderboard_id}

    # TODO: Implement POST /ranking/request/action/rt/create
    # TODO: Implement POST /ranking/request/action/rt/vote
    # TODO: Implement POST /ranking/request/action/rt/comment

    # TODO: Implement POST /ranking/request/action/qat/vote
    # TODO: Implement POST /ranking/request/action/qat/comment

    # TODO: Implement POST /ranking/request/action/nat/replace
    # TODO: Implement POST /ranking/request/action/nat/qualify
    # TODO: Implement POST /ranking/request/action/nat/deny



from dataclasses import dataclass
from typing import List

from dataclasses_json import dataclass_json, LetterCase

from .metadata import Metadata
from .score import Score


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ScoreCollection:
    scores: List[Score]
    metadata: Metadata



import asyncio
import json
import logging
import typing
from asyncio import AbstractEventLoop
from datetime import datetime
from enum import Enum
from typing import Optional, AsyncIterable

import aiohttp
from aiohttp import ClientResponse, ClientResponseError, ClientSession, ClientWebSocketResponse, WSMessage

from .errors import ScoreSaberException, NotFoundException
from .version import __version__

_logger = logging.getLogger(__name__)


class HttpClient:
    _ws_url = "ws://scoresaber.com/ws"
    _headers = {
        'User-Agent': f'PyScoreSaber/{__version__}'
    }
    MAX_TIMEOUT = 60

    def __init__(self, loop: Optional[AbstractEventLoop] = None):
        self.loop = loop
        self._aiohttp: Optional[ClientSession] = None
        self._ws: Optional[ClientWebSocketResponse] = None

    async def start(self):
        if self._aiohttp is None:
            self._aiohttp = aiohttp.ClientSession(loop=self.loop, headers=self._headers, raise_for_status=True)

    async def close(self):
        if self._aiohttp is not None:
            await self._aiohttp.close()
            self._aiohttp = None

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def _request(self, *args, **kwargs) -> ClientResponse:
        retries = 0

        while True:
            try:
                response = await self._aiohttp.request(*args, **kwargs)

                if response.status == 200:
                    return response

                raise ScoreSaberException(response.status, str(response.real_url))
            except ClientResponseError as error:
                status = error.status
                real_url = str(error.request_info.real_url)

                if status == 404:
                    raise NotFoundException(status, real_url) from error

            sleep = 10 * retries

            if sleep > self.MAX_TIMEOUT:
                sleep = 60

            _logger.warning(f"[{status}] Request failed! Waiting {sleep} seconds...")
            await asyncio.sleep(sleep)

            retries += 1

    async def ws_connect(self):
        if self._ws is None:
            self._ws = await self._aiohttp.ws_connect(self._ws_url)

    async def ws_close(self):
        if self._ws is not None:
            await self._ws.close()
            self._ws = None

    async def ws_listen(self) -> AsyncIterable[typing.Dict]:
        if self._ws is None:
            await self.ws_connect()

        async for message in self._ws:
            if message.type == aiohttp.WSMsgType.CLOSE:
                _logger.warning(f"Websocket closed! Reconnecting...")
                await self.ws_connect()
            if message.type == aiohttp.WSMsgType.TEXT:
                text = message.data

                if text == "Connected to the ScoreSaber WSS":
                    continue

            yield message.json()

    def _format_params(self, params):
        for key, value in params.copy().items():
            if value is None:
                del params[key]
                continue

            params[key] = self._format_value(value)

        return params

    @staticmethod
    def _format_value(value):
        if isinstance(value, datetime):
            return 1000 * int(value.timestamp())

        if isinstance(value, Enum):
            return value.value

        return value

    async def get(self, type_, url, params={}, *args, **kwargs):
        params = self._format_params(params)

        response = await self._request('GET', url, params=params, *args, **kwargs)
        data = await response.json()

        if not typing.get_args(type_):
            if isinstance(data, int):
                return data

            return type_.from_dict(data)

        return typing.get_args(type_)[0].schema().load(data, many=True)



from dataclasses import dataclass
from typing import List

from dataclasses_json import dataclass_json, LetterCase

from .metadata import Metadata
from .player_score import PlayerScore


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlayerScoresCollection:
    """Player scores collection data from ScoreSaber"""

    player_scores: List[PlayerScore]
    metadata: Metadata



from dataclasses import dataclass
from typing import Optional

from dataclasses_json import dataclass_json

from .fields import default


@dataclass_json
@dataclass
class LeaderBoardPlayer:
    """Leaderboard player data from ScoreSaber"""

    id: str = default()
    name: str = default()
    profile_picture: str = default("profilePicture")
    country: str = default()
    permissions: int = default()
    role: Optional[str] = default()



import math
from typing import *

from .scoresaber import ScoreSaber
from .models import *


class ScoreSaberAPI(ScoreSaber):
    async def players_basic(self, player_ids: List[int]) -> AsyncIterable[Player]:
        for player_id in player_ids:
            yield await self.player_basic(player_id)

    async def players_full(self, player_ids: List[int]) -> AsyncIterable[Player]:
        for player_id in player_ids:
            yield await self.player_full(player_id)

    async def player_scores_all(self, player_id: int, score_sort: ScoreSort) -> AsyncIterable[List[PlayerScore]]:
        page = 1
        max_page = -1

        while page < max_page or max_page == -1:
            recent_scores = await self.player_scores(player_id, sort=score_sort, limit=100, page=page)

            if max_page == -1:
                max_page = math.ceil(recent_scores.metadata.total / 100) + 1

            yield recent_scores.player_scores

            page += 1



from dataclasses import dataclass
from typing import List

from dataclasses_json import dataclass_json

from .fields import default
from .leaderboard_info import LeaderboardInfo
from .metadata import Metadata


@dataclass_json
@dataclass
class LeaderboardInfoCollection:
    """Leaderboard info collection data from ScoreSaber"""

    leaderboards: List[LeaderboardInfo] = default()
    metadata: Metadata = default()



from dataclasses import dataclass

from dataclasses_json import dataclass_json, LetterCase

from .enum import BeatmapDifficulty, GameMode
from .fields import difficulty_field, game_mode_field


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Difficulty:
    """Difficulty data from ScoreSaber"""

    leaderboard_id: int
    difficulty_raw: str
    game_mode: GameMode = game_mode_field("gameMode")
    difficulty: BeatmapDifficulty = difficulty_field()



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from dataclasses_json import dataclass_json

from .fields import datetime_field, default
from .leaderboard_player import LeaderBoardPlayer


@dataclass_json
@dataclass
class Score:
    """Score data from ScoreSaber"""

    id: int = default()
    leaderboard_player_info: Optional[LeaderBoardPlayer] = default("leaderboardPlayerInfo")
    rank: int = default()
    base_score: int = default("baseScore")
    modified_score: int = default("modifiedScore")
    pp: float = default()
    weight: float = default()
    modifiers: str = default()
    multiplier: int = default()
    bad_cuts: int = default("badCuts")
    missed_notes: int = default("missedNotes")
    max_combo: int = default("maxCombo")
    full_combo: bool = default("fullCombo")
    hmd: int = default()
    has_replay: bool = default("hasReplay")
    time_set: datetime = datetime_field("timeSet")



import logging

from .scoresaber import ScoreSaber
from .scoresaber_api import ScoreSaberAPI
from .scoresaber_provider import ScoreSaberProvider
from .errors import *
from .models import *
from .version import __version__

logging.getLogger("pyscoresaber").addHandler(logging.NullHandler())



from typing import Optional

import setuptools

# Get version info
__version__: Optional[str] = None
exec(open('src/pyscoresaber/version.py').read())

if __version__ is None:
    raise RuntimeError("Failed to get version!")

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Requirements
requirements = []
with open("requirements.txt", "r", encoding="utf-8") as f:
    requirements += f.read().splitlines()

extras_require = {
    'test': [
        'coverage',
        'pytest',
        'pytest-asyncio',
        'pytest-cov'
    ]
}

setuptools.setup(
    name="PyScoreSaber",
    version=__version__,
    author="LuCkEr-",
    author_email="lucker@lucker.xyz",
    description="Score Saber API wrapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Kiyomi-Parents/PyScoreSaber",
    project_urls={
        "Bug Tracker": "https://github.com/Kiyomi-Parents/PyScoreSaber/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
    install_requires=requirements,
    extras_require=extras_require,
)



from .enum import *

from .badge import Badge
from .difficulty import Difficulty
from .leaderboard_info import LeaderboardInfo
from .leaderboard_info_collection import LeaderboardInfoCollection
from .leaderboard_player import LeaderBoardPlayer
from .metadata import Metadata
from .player import Player
from .player_collection import PlayerCollection
from .player_score import PlayerScore
from .player_scores_collection import PlayerScoresCollection
from .score import Score
from .score_collection import ScoreCollection
from .score_stats import ScoreStats



from dataclasses import dataclass
from typing import List

from dataclasses_json import dataclass_json, LetterCase

from .metadata import Metadata
from .player import Player


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlayerCollection:
    """Player collection data from ScoreSaber"""

    players: List[Player]
    metadata: Metadata



from dataclasses import field, Field
from datetime import datetime
from typing import Optional

from dataclasses_json import config
from dateutil import parser
from marshmallow import fields

from .enum import GameMode, BeatmapDifficulty


def datetime_from_iso_format(time):
    if time:
        return parser.isoparse(time)

    return None


def datetime_field(json_field_name: Optional[str] = None) -> Field:
    if json_field_name is None:
        conf = config(
                encoder=datetime.isoformat,
                decoder=datetime_from_iso_format,
                mm_field=fields.DateTime(format='iso')
            )
    else:
        conf = config(
            encoder=datetime.isoformat,
            decoder=datetime_from_iso_format,
            mm_field=fields.DateTime(format='iso'),
            field_name=json_field_name,
        )

    return field(
        default=None,
        metadata=conf
    )


def game_mode_decoder(value: any) -> GameMode:
    if GameMode.has_value(value):
        return GameMode(value)

    # Some weird "StandardHM" characteristic that doesnt exist anymore
    if "Standard" in value:
        return GameMode.STANDARD

    if GameMode.has_value(value):
        return GameMode(value)

    return GameMode.UNKNOWN


def game_mode_encoder(game_mode: GameMode) -> str:
    return game_mode.value


def game_mode_field(json_field_name: Optional[str] = None) -> Field:
    return field(
        default=None,
        metadata=config(
            encoder=game_mode_encoder,
            decoder=game_mode_decoder,
            field_name=json_field_name
        )
    )


def difficulty_decoder(value: any) -> BeatmapDifficulty:
    if BeatmapDifficulty.has_value(value):
        return BeatmapDifficulty(value)

    return BeatmapDifficulty.UNKNOWN


def difficulty_encoder(beatmap_difficulty: BeatmapDifficulty) -> int:
    return beatmap_difficulty.value


def difficulty_field(json_field_name: Optional[str] = None) -> Field:
    return field(
        default=None,
        metadata=config(
            encoder=difficulty_encoder,
            decoder=difficulty_decoder,
            field_name=json_field_name
        )
    )


def default(json_field_name: Optional[str] = None) -> Field:
    if json_field_name is None:
        conf = config()
    else:
        conf = config(field_name=json_field_name)

    return field(
        default=None,
        metadata=conf
    )
